use clap::{FromArgMatches, Subcommand};
use microrm::prelude::*;
use microrm::schema::datum::{Datum, DatumDiscriminatorRef};
use microrm::schema::entity::{Entity, EntityPartList, EntityPartVisitor, EntityID};

use crate::UIDCError;
use crate::schema::UIDCDatabase;

mod role;
mod user;

pub trait Object: Sized + Entity + std::fmt::Debug {
    type CreateParameters : clap::Parser + std::fmt::Debug;
    fn create_from_params(_: &Self::CreateParameters) -> Result<Self, UIDCError>;
    fn extra_commands() -> impl Iterator<Item = clap::Command> { vec![].into_iter() }

    /// get the relevant IDMap from the database
    fn db_object(db: &UIDCDatabase) -> &IDMap<Self> where Self: Sized;
    fn build_uniques(strings: &Vec<String>) -> <Self::Uniques as EntityPartList>::DatumList;

    fn shortname(&self) -> &str;
}

pub trait ObjectExt: Sized + Object {
    fn create(ctx: &impl Insertable<Self>, cp: &Self::CreateParameters) -> Result<(), UIDCError> {
        ctx.insert(Self::create_from_params(cp)?)?;
        Ok(())
    }

    fn delete(ctx: impl microrm::prelude::Queryable<EntityOutput = Self>,
               which: <Self::Uniques as EntityPartList>::DatumList)
        -> Result<(), UIDCError> {
        ctx.unique(which).delete()?;
        Ok(())
    }

    fn list_all(ctx: impl microrm::prelude::Queryable<EntityOutput = Self>) -> Result<(), UIDCError> {
        println!("Listing all {}(s): ({})", Self::entity_name(), ctx.clone().count()?);
        for obj in ctx.get()?.into_iter() {
            println!(" - {}", obj.shortname());
        }

        Ok(())
    }

    fn inspect(ctx: impl microrm::prelude::Queryable<EntityOutput = Self>,
               which: <Self::Uniques as EntityPartList>::DatumList) -> Result<(), UIDCError> {

        let obj = ctx.unique(which).get()?.ok_or(UIDCError::Abort("no such element, cannot inspect"))?;
        println!("{:#?}", obj.as_ref());

        fn inspect_ai<AI: AssocInterface>(name: &'static str, ai: &AI) {
            println!("{}: ({})", name, ai.count().unwrap());
            for a in ai.get().expect("couldn't get object associations") {
                println!("[#{:3}]: {:?}", a.id().into_raw(), a.wrapped());
            }
        }

        struct AssocFieldWalker;
        impl EntityPartVisitor for AssocFieldWalker {
            fn visit_datum<EP: microrm::schema::entity::EntityPart>(&mut self, datum: &EP::Datum) {
                struct Discriminator<'l, D: Datum>(&'l D, &'static str);

                impl<'l, D: Datum> DatumDiscriminatorRef for Discriminator<'l, D> {
                    fn visit_serialized<T: serde::Serialize + serde::de::DeserializeOwned>(&mut self, _: &T) {}
                    fn visit_bare_field<T: Datum>(&mut self, _: &T) {}
                    fn visit_entity_id<E: Entity>(&mut self, _: &E::ID) {}
                    fn visit_assoc_map<E: Entity>(&mut self, amap: &AssocMap<E>) {
                        inspect_ai(self.1, amap);
                    }
                    fn visit_assoc_domain<R: microrm::schema::Relation>(&mut self, adomain: &microrm::schema::AssocDomain<R>) {
                        inspect_ai(self.1, adomain);
                    }
                    fn visit_assoc_range<R: microrm::schema::Relation>(&mut self, arange: &microrm::schema::AssocRange<R>) {
                        inspect_ai(self.1, arange);
                    }
                }

                datum.accept_discriminator_ref(&mut Discriminator(datum, EP::part_name()));
            }
        }

        obj.accept_part_visitor_ref(&mut AssocFieldWalker);

        Ok(())
    }
}

impl<T: Object + std::fmt::Debug> ObjectExt for T {}

#[derive(Clone, Debug)]
enum InterfaceVerb<O: ObjectExt> {
    Attach,
    Create(O::CreateParameters),
    Delete(Vec<String>),
    Detach,
    ListAll,
    Inspect(Vec<String>),
}

impl<O: ObjectExt> InterfaceVerb<O> {
    fn from_matches(matches: &clap::ArgMatches) -> Result<Self, clap::Error> {
        let (subcommand, matches) =
            matches.subcommand().ok_or(clap::Error::new(clap::error::ErrorKind::MissingSubcommand)).unwrap();

        let parse_uniques = || {
            struct UVisitor<'a>(&'a clap::ArgMatches, &'a mut Vec<String>);
            impl<'a> EntityPartVisitor for UVisitor<'a> {
                fn visit<EP: microrm::schema::entity::EntityPart>(&mut self) {
                    self.1.push(self.0.get_one::<std::string::String>(EP::part_name()).unwrap().clone());
                }
            }

            let mut unique_values = vec![];
            <O::Uniques as EntityPartList>::accept_part_visitor(&mut UVisitor(matches, &mut unique_values));
            unique_values
        };

        Ok(match subcommand {
            "attach" => InterfaceVerb::Attach,
            "create" => InterfaceVerb::Create(
                <O::CreateParameters as clap::FromArgMatches>::from_arg_matches(matches)?
            ),
            "delete" => InterfaceVerb::Delete(parse_uniques()),
            "detach" => InterfaceVerb::Detach,
            "list" => InterfaceVerb::ListAll,
            "inspect" => InterfaceVerb::Inspect(parse_uniques()),
            _ => unreachable!()
        })
    }
}

#[derive(Debug)]
pub struct ClapInterface<O: ObjectExt> {
    verb: InterfaceVerb<O>,
    // uniques: Vec<String>,
    _ghost: std::marker::PhantomData<O>,
}

impl<O: ObjectExt> ClapInterface<O> {
    pub fn perform(&self, query_ctx: impl microrm::prelude::Queryable::<EntityOutput = O>, insert_ctx: &impl microrm::prelude::Insertable<O>) 
        -> Result<(), UIDCError> {
        match &self.verb {
            InterfaceVerb::Attach => {
                todo!()
            },
            InterfaceVerb::Create(params) => {
                O::create(insert_ctx, &params)?;
            },
            InterfaceVerb::Delete(uniques) => {
                O::delete(query_ctx, O::build_uniques(uniques))?;
            },
            InterfaceVerb::Detach => {
                todo!()
            },
            InterfaceVerb::ListAll => {
                O::list_all(query_ctx)?;
            },
            InterfaceVerb::Inspect(uniques) => {
                O::inspect(query_ctx, O::build_uniques(uniques))?;
            },
        }
        Ok(())
    }

    /// iterate across the list of unique parts (O::Uniques) and add args for each
    fn add_uniques(mut cmd: clap::Command) -> clap::Command {
        struct UVisitor<'a>(&'a mut clap::Command);
        impl<'a> EntityPartVisitor for UVisitor<'a> {
            fn visit<EP: microrm::schema::entity::EntityPart>(&mut self) {
                let arg = clap::Arg::new(EP::part_name())
                    .required(true)
                    .help(EP::desc());
                *self.0 = self.0.clone().arg(arg);
            }
        }

        <O::Uniques as EntityPartList>::accept_part_visitor(&mut UVisitor(&mut cmd));

        cmd
    }
}

impl<O: ObjectExt> FromArgMatches for ClapInterface<O> {
    fn from_arg_matches(matches: &clap::ArgMatches) -> Result<Self, clap::Error> {
        let verb = InterfaceVerb::from_matches(matches);

        Ok(Self {
            verb: verb?,
            _ghost: Default::default(),
        })
    }

    fn update_from_arg_matches(&mut self, matches: &clap::ArgMatches) -> Result<(), clap::Error> {
        todo!()
    }
}

impl<O: ObjectExt> Subcommand for ClapInterface<O> {
    fn has_subcommand(name: &str) -> bool {
        todo!()
    }

    fn augment_subcommands(cmd: clap::Command) -> clap::Command {
        cmd
            .subcommand(<O::CreateParameters as clap::CommandFactory>::command().name("create"))
            .subcommand(Self::add_uniques(clap::Command::new("delete")))
            .subcommand(Self::add_uniques(clap::Command::new("inspect")))
            .subcommand(clap::Command::new("list"))
            .subcommands(O::extra_commands())
    }

    fn augment_subcommands_for_update(cmd: clap::Command) -> clap::Command {
        todo!()
    }
}
